@@ -556,9 +556,19 @@ void hcid_dbus_inquiry_complete(bdaddr_t *local)
 	if (found_device_req_name(adapter) == 0)
 		return;
 
+	state = adapter_get_state(adapter);
+
+	/*
+	 * workaround to identify situation when there is no devices around
+	 * but periodic inquiry is active.
+	 */
+	if (!(state & STD_INQUIRY) && !(state & PERIODIC_INQUIRY)) {
+		state |= PERIODIC_INQUIRY;
+		adapter_set_state(adapter, state);
+	}
+
 	/* reset the discover type to be able to handle D-Bus and non D-Bus
 	 * requests */
-	state = adapter_get_state(adapter);
 	state &= ~STD_INQUIRY;
 	state &= ~PERIODIC_INQUIRY;
 	adapter_set_state(adapter, state);
@@ -661,17 +671,20 @@ void hcid_dbus_inquiry_result(bdaddr_t *local, bdaddr_t *peer, uint32_t class,
 		state |= PERIODIC_INQUIRY;
 		adapter_set_state(adapter, state);
 	}
-	/* Out of range list update */
-	if (adapter_get_state(adapter) & PERIODIC_INQUIRY)
-		adapter_remove_oor_device(adapter, peer_addr);
+
+	legacy = (data == NULL);
 
 	memset(&match, 0, sizeof(struct remote_dev_info));
 	bacpy(&match.bdaddr, peer);
 	match.name_status = NAME_SENT;
 	/* if found: don't send the name again */
 	dev = adapter_search_found_devices(adapter, &match);
-	if (dev)
+	if (dev) {
+		adapter_update_found_devices(adapter, peer, rssi, class,
+						NULL, NULL, legacy,
+						NAME_NOT_REQUIRED);
 		return;
+	}
 
 	/* the inquiry result can be triggered by NON D-Bus client */
 	if (adapter_get_state(adapter) & RESOLVE_NAME)
@@ -703,7 +716,6 @@ void hcid_dbus_inquiry_result(bdaddr_t *local, bdaddr_t *peer, uint32_t class,
 		}
 	}
 
-	legacy = (data == NULL);
 
 	if (name && name_type != 0x08)
 		name_status = NAME_SENT;
@@ -792,6 +804,7 @@ int hcid_dbus_link_key_notify(bdaddr_t *local, bdaddr_t *peer,
 	struct btd_device *device;
 	struct btd_adapter *adapter;
 	uint8_t local_auth = 0xff, remote_auth, new_key_type;
+	gboolean bonding;
 
 	if (!get_adapter_and_device(local, peer, &adapter, &device, TRUE))
 		return -ENODEV;
@@ -803,6 +816,7 @@ int hcid_dbus_link_key_notify(bdaddr_t *local, bdaddr_t *peer,
 
 	get_auth_requirements(local, peer, &local_auth);
 	remote_auth = device_get_auth(device);
+	bonding = device_is_bonding(device, NULL);
 
 	debug("local auth 0x%02x and remote auth 0x%02x",
 					local_auth, remote_auth);
@@ -831,7 +845,10 @@ int hcid_dbus_link_key_notify(bdaddr_t *local, bdaddr_t *peer,
 			return err;
 		}
 
-		device_set_temporary(device, FALSE);
+		/* If not the initiator consider the device permanent otherwise
+		 * wait to service discover to complete */
+		if (!bonding)
+			device_set_temporary(device, FALSE);
 	}
 
 	/* If this is not the first link key set a flag so a subsequent auth
@@ -841,7 +858,7 @@ int hcid_dbus_link_key_notify(bdaddr_t *local, bdaddr_t *peer,
 
 	if (!device_is_connected(device))
 		device_set_secmode3_conn(device, TRUE);
-	else if (!device_is_bonding(device, NULL) && old_key_type == 0xff)
+	else if (!bonding && old_key_type == 0xff)
 		hcid_dbus_bonding_process_complete(local, peer, 0);
 
 	return 0;
