@@ -57,8 +57,6 @@
 
 struct main_opts main_opts;
 
-static gboolean starting = TRUE;
-
 static GKeyFile *load_config(const char *file)
 {
 	GError *err = NULL;
@@ -150,8 +148,8 @@ static void parse_config(GKeyFile *config)
 		debug("%s", err->message);
 		g_clear_error(&err);
 	} else {
-		debug("inqmode=%d", val);
-		main_opts.inqmode = val;
+		debug("discov_interval=%d", val);
+		main_opts.discov_interval = val;
 	}
 
 	boolean = g_key_file_get_boolean(config, "General",
@@ -226,14 +224,11 @@ static void update_service_classes(const bdaddr_t *bdaddr, uint8_t value)
 		if (hci_test_bit(HCI_RAW, &di.flags))
 			continue;
 
-		if (!hci_test_bit(HCI_UP, &di.flags))
-			continue;
-
 		if (bacmp(bdaddr, BDADDR_ANY) != 0 &&
 				bacmp(bdaddr, &di.bdaddr) != 0)
 			continue;
 
-		manager_update_adapter(di.dev_id, value, starting);
+		manager_update_adapter(di.dev_id, value);
 	}
 
 	g_free(dl);
@@ -417,8 +412,6 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
-	starting = FALSE;
-
 	manager_startup_complete();
 
 	debug("Entering main loop");
